@@ -23,8 +23,7 @@ use constant REPORT            => $ENV{JSON_VALIDATOR_REPORT} // DEBUG >= 2;
 use constant RECURSION_LIMIT   => $ENV{JSON_VALIDATOR_RECURSION_LIMIT} || 100;
 use constant SPECIFICATION_URL => 'http://json-schema.org/draft-04/schema#';
 
-our $DEFINITIONS = 'definitions';
-our $VERSION = '3.14';
+our $VERSION = '3.15';
 our $YAML_LOADER = eval q[use YAML::XS 0.67; YAML::XS->can('Load')];  # internal
 our @EXPORT_OK   = qw(joi validate_json);
 
@@ -60,50 +59,54 @@ has ua => sub {
 
 sub bundle {
   my ($self, $args) = @_;
-  my @topics = ([undef, my $bundle = {}]);
   my ($cloner, $tied);
 
-  $topics[0][0]
+  my $schema
     = $args->{schema} ? $self->_resolve($args->{schema}) : $self->schema->data;
-
-  local $DEFINITIONS = $args->{ref_key} || $DEFINITIONS;
-  Mojo::Util::deprecated('bundle({ref_key => "..."}) will be removed.')
-    if $args->{ref_key};
+  my @topics = ([$schema, my $bundle = {}, '']);    # ([$from, $to], ...);
 
   if ($args->{replace}) {
-    $cloner = sub {
-      my $from = shift;
-      my $ref  = ref $from;
-      $from = $tied->schema if $ref eq 'HASH' and $tied = tied %$from;
-      my $to = $ref eq 'ARRAY' ? [] : $ref eq 'HASH' ? {} : $from;
-      push @topics, [$from, $to] if $ref;
-      return $to;
-    };
-  }
-  else {
-    my %seen;
-    $bundle->{$DEFINITIONS} = $topics[0][0]{$DEFINITIONS} || {};
     $cloner = sub {
       my $from      = shift;
       my $from_type = ref $from;
-
-      if ($from_type eq 'HASH' and my $ref = tied %$from) {
-        return $from
-          if !$args->{schema}
-          and $ref->fqn =~ m!^\Q$self->{root_schema_url}\E\#!;
-
-        my $k = $self->_definitions_key($bundle, $ref, \%seen);
-        push @topics, [$ref->schema, $bundle->{$DEFINITIONS}{$k} ||= {}]
-          unless $seen{$ref->fqn}++;
-        tie my %ref, 'JSON::Validator::Ref', $ref->schema, "#/$DEFINITIONS/$k";
-        return \%ref;
-      }
-
+      $from = $tied->schema if $from_type eq 'HASH' and $tied = tied %$from;
       my $to = $from_type eq 'ARRAY' ? [] : $from_type eq 'HASH' ? {} : $from;
       push @topics, [$from, $to] if $from_type;
       return $to;
     };
   }
+  else {
+    $cloner = sub {
+      my $from      = shift;
+      my $from_type = ref $from;
+
+      $tied = $from_type eq 'HASH' && tied %$from;
+      unless ($tied) {
+        my $to = $from_type eq 'ARRAY' ? [] : $from_type eq 'HASH' ? {} : $from;
+        push @topics, [$from, $to] if $from_type;
+        return $to;
+      }
+
+      return $from
+        if !$args->{schema}
+        and $tied->fqn =~ m!^\Q$self->{root_schema_url}\E\#!;
+
+      my $p = $self->_definitions_path($bundle, $tied);
+      unless ($self->{bundled_refs}{$tied->fqn}) {
+        $self->{bundled_refs}{$tied->fqn} = $tied;
+        push @topics, [$schema->{$p->[0]} || {}, $bundle->{$p->[0]} ||= {}];
+        push @topics, [$tied->schema, $bundle->{$p->[0]}{$p->[1]} ||= {}];
+      }
+
+      tie my %ref, 'JSON::Validator::Ref', $tied->schema, "#/$p->[0]/$p->[1]";
+      return \%ref;
+    };
+  }
+
+  Mojo::Util::deprecated('bundle({ref_key => "..."}) will be removed.')
+    if $args->{ref_key};
+  local $self->{definitions_key} = $args->{ref_key};
+  local $self->{bundled_refs}    = {};
 
   while (@topics) {
     my ($from, $to) = @{shift @topics};
@@ -119,7 +122,6 @@ sub bundle {
     }
   }
 
-  delete $bundle->{$DEFINITIONS} unless keys %{$bundle->{$DEFINITIONS}};
   return $bundle;
 }
 
@@ -225,27 +227,34 @@ sub _build_formats {
   };
 }
 
-sub _definitions_key {
-  my ($self, $bundle, $ref, $seen) = @_;
+sub _definitions_path {
+  my ($self, $bundle, $ref) = @_;
+  my $definitions_key = $self->{definitions_key} || 'definitions';
 
-  # No need to rewrite, when it already has a nice name
-  return $1
-    if $ref->fqn =~ m!#/$DEFINITIONS/([^/]+)$!
-    and ($seen->{$ref->fqn}
-    or !$bundle->{$DEFINITIONS}{$1}
-    or D($ref->schema) eq D($bundle->{$DEFINITIONS}{$1}));
+  # No need to rewrite, if it already has a nice name
+  if ($ref->fqn =~ m!#/$definitions_key/([^/]+)$!) {
+    my $key = $1;
 
-  # Must mask path to file on disk
-  my $key       = $ref->fqn;
-  my $spec_path = (split '#', $key)[0];
+    if ( $self->{bundled_refs}{$ref->fqn}
+      or !$bundle->{$definitions_key}{$key}
+      or D($ref->schema) eq D($bundle->{$definitions_key}{$key}))
+    {
+      return [$definitions_key, $key];
+    }
+  }
+
+  # Generate definitions key based on filename
+  my ($spec_path, $fragment) = split '#', $ref->fqn;
+  my $key = $fragment;
   if (-e $spec_path) {
-    $key = sprintf '%s-%s', substr(sha1_sum($key), 0, 10),
-      path($spec_path)->basename;
+    $key = join '-', map { s!^\W+!!; $_ } grep {$_} path($spec_path)->basename,
+      $fragment, substr(sha1_sum($spec_path), 0, 10);
   }
 
   # Fallback or nicer path name
   $key =~ s![^\w-]!_!g;
-  $key;
+
+  return [$definitions_key, $key];
 }
 
 sub _get {
@@ -611,7 +620,7 @@ sub _validate {
   if (my $rules = $schema->{not}) {
     push @errors, $self->_validate($to_json ? $$to_json : $_[1], $path, $rules);
     $self->_report_errors($path, 'not', \@errors) if REPORT;
-    return @errors ? () : (E $path, 'Should not match.');
+    return @errors ? () : (E $path, [not => 'not']);
   }
 
   if (my $rules = $schema->{allOf}) {
@@ -650,7 +659,7 @@ sub _validate_all_of {
   }
 
   $self->_report_errors($path, 'allOf', \@errors) if REPORT;
-  return E $path, "/allOf Expected @{[join '/', _uniq(@expected)]} - got $type."
+  return E $path, [allOf => type => join('/', _uniq(@expected)), $type]
     if !@errors and @expected;
   return _add_path_to_error_messages(allOf => @errors) if @errors;
   return;
@@ -678,7 +687,7 @@ sub _validate_any_of {
 
   $self->_report_errors($path, 'anyOf', \@errors) if REPORT;
   my $expected = join '/', _uniq(@expected);
-  return E $path, "/anyOf Expected $expected - got $type." unless @errors;
+  return E $path, [anyOf => type => $expected, $type] unless @errors;
   return _add_path_to_error_messages(anyOf => @errors);
 }
 
@@ -711,8 +720,8 @@ sub _validate_one_of {
 
   return if @errors + @expected + 1 == @$rules;
   my $expected = join '/', _uniq(@expected);
-  return E $path, "All of the oneOf rules match." unless @errors + @expected;
-  return E $path, "/oneOf Expected $expected - got $type." unless @errors;
+  return E $path, [oneOf => 'all_rules_match'] unless @errors + @expected;
+  return E $path, [oneOf => type => $expected => $type] unless @errors;
   return _add_path_to_error_messages(oneOf => @errors);
 }
 
@@ -725,9 +734,9 @@ sub _validate_type_enum {
     return if $m eq S $i;
   }
 
-  local $" = ', ';
-  return E $path, sprintf 'Not in enum list: %s.', join ', ',
+  $enum = join ', ',
     map { (!defined or ref) ? Mojo::JSON::encode_json($_) : $_ } @$enum;
+  return E $path, [enum => enum => $enum];
 }
 
 sub _validate_type_const {
@@ -736,8 +745,7 @@ sub _validate_type_const {
   my $m     = S $data;
 
   return if $m eq S $const;
-  return E $path, sprintf 'Does not match const: %s.',
-    Mojo::JSON::encode_json($const);
+  return E $path, [const => const => Mojo::JSON::encode_json($const)];
 }
 
 sub _validate_format {
@@ -746,7 +754,7 @@ sub _validate_format {
   return do { warn "Format rule for '$schema->{format}' is missing"; return }
     unless $code;
   return unless my $err = $code->($value);
-  return E $path, $err;
+  return E $path, [format => $schema->{format}, $err];
 }
 
 sub _validate_type_any { }
@@ -756,21 +764,21 @@ sub _validate_type_array {
   my @errors;
 
   if (ref $data ne 'ARRAY') {
-    return E $path, _expected(array => $data);
+    return E $path, [array => type => _guess_data_type($data)];
   }
   if (defined $schema->{minItems} and $schema->{minItems} > @$data) {
-    push @errors, E $path, sprintf 'Not enough items: %s/%s.', int @$data,
-      $schema->{minItems};
+    push @errors, E $path,
+      [array => minItems => int(@$data), $schema->{minItems}];
   }
   if (defined $schema->{maxItems} and $schema->{maxItems} < @$data) {
-    push @errors, E $path, sprintf 'Too many items: %s/%s.', int @$data,
-      $schema->{maxItems};
+    push @errors, E $path,
+      [array => maxItems => int(@$data), $schema->{maxItems}];
   }
   if ($schema->{uniqueItems}) {
     my %uniq;
     for (@$data) {
       next if !$uniq{S($_)}++;
-      push @errors, E $path, 'Unique items required.';
+      push @errors, E $path, [array => 'uniqueItems'];
       last;
     }
   }
@@ -797,8 +805,8 @@ sub _validate_type_array {
       }
     }
     elsif (!$additional_items) {
-      push @errors, E $path, sprintf "Invalid number of items: %s/%s.",
-        int(@$data), int(@rules);
+      push @errors, E $path,
+        [array => additionalItems => int(@$data), int(@rules)];
     }
   }
   elsif (UNIVERSAL::isa($schema->{items}, 'HASH')) {
@@ -832,7 +840,7 @@ sub _validate_type_boolean {
     return;
   }
 
-  return E $path, _expected(boolean => $value);
+  return E $path, [boolean => type => _guess_data_type($value)];
 }
 
 sub _validate_type_integer {
@@ -841,14 +849,14 @@ sub _validate_type_integer {
 
   return @errors if @errors;
   return if $value =~ /^-?\d+$/;
-  return E $path, "Expected integer - got number.";
+  return E $path, [integer => type => _guess_data_type($value)];
 }
 
 sub _validate_type_null {
   my ($self, $value, $path, $schema) = @_;
 
-  return E $path, 'Not null.' if defined $value;
-  return;
+  return unless defined $value;
+  return E $path, ['null', 'null'];
 }
 
 sub _validate_type_number {
@@ -858,10 +866,10 @@ sub _validate_type_number {
   $expected ||= 'number';
 
   if (!defined $value or ref $value) {
-    return E $path, _expected($expected => $value);
+    return E $path, [$expected => type => _guess_data_type($value)];
   }
   unless (_is_number($value)) {
-    return E $path, "Expected $expected - got string."
+    return E $path, [$expected => type => _guess_data_type($value)]
       if !$self->{coerce}{numbers}
       or $value !~ /^-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?$/;
     $_[1] = 0 + $value;    # coerce input value
@@ -873,16 +881,16 @@ sub _validate_type_number {
   if (my $e
     = _cmp($schema->{minimum}, $value, $schema->{exclusiveMinimum}, '<'))
   {
-    push @errors, E $path, "$value $e minimum($schema->{minimum})";
+    push @errors, E $path, [$expected => minimum => $value, $schema->{minimum}];
   }
   if (my $e
     = _cmp($value, $schema->{maximum}, $schema->{exclusiveMaximum}, '>'))
   {
-    push @errors, E $path, "$value $e maximum($schema->{maximum})";
+    push @errors, E $path, [$expected => maximum => $value, $schema->{maximum}];
   }
   if (my $d = $schema->{multipleOf}) {
     if (($value / $d) =~ /\.[^0]+$/) {
-      push @errors, E $path, "Not multiple of $d.";
+      push @errors, E $path, [$expected => multipleOf => $d];
     }
   }
 
@@ -895,17 +903,17 @@ sub _validate_type_object {
   my ($additional, @errors, %rules);
 
   if (ref $data ne 'HASH') {
-    return E $path, _expected(object => $data);
+    return E $path, [object => type => _guess_data_type($data)];
   }
 
   my @dkeys = sort keys %$data;
   if (defined $schema->{maxProperties} and $schema->{maxProperties} < @dkeys) {
-    push @errors, E $path, sprintf 'Too many properties: %s/%s.', int @dkeys,
-      $schema->{maxProperties};
+    push @errors, E $path,
+      [object => maxProperties => int(@dkeys), $schema->{maxProperties}];
   }
   if (defined $schema->{minProperties} and $schema->{minProperties} > @dkeys) {
-    push @errors, E $path, sprintf 'Not enough properties: %s/%s.', int @dkeys,
-      $schema->{minProperties};
+    push @errors, E $path,
+      [object => minProperties => int(@dkeys), $schema->{minProperties}];
   }
   if (my $n_schema = $schema->{propertyNames}) {
     for my $name (keys %$data) {
@@ -942,12 +950,12 @@ sub _validate_type_object {
   }
   elsif (my @k = grep { !$rules{$_} } @dkeys) {
     local $" = ', ';
-    return E $path, "Properties not allowed: @k.";
+    return E $path, [object => additionalProperties => join '/', @k];
   }
 
   for my $k (sort keys %required) {
     next if exists $data->{$k};
-    push @errors, E _path($path, $k), 'Missing property.';
+    push @errors, E _path($path, $k), [object => 'required'];
     delete $rules{$k};
   }
 
@@ -974,13 +982,13 @@ sub _validate_type_string {
   my @errors;
 
   if (!defined $value or ref $value) {
-    return E $path, _expected(string => $value);
+    return E $path, [string => type => _guess_data_type($value)];
   }
   if (  B::svref_2object(\$value)->FLAGS & (B::SVp_IOK | B::SVp_NOK)
     and 0 + $value eq $value
     and $value * 0 == 0)
   {
-    return E $path, "Expected string - got number."
+    return E $path, [string => type => _guess_data_type($value)]
       unless $self->{coerce}{strings};
     $_[1] = "$value";    # coerce input value
   }
@@ -989,21 +997,19 @@ sub _validate_type_string {
   }
   if (defined $schema->{maxLength}) {
     if (length($value) > $schema->{maxLength}) {
-      push @errors, E $path, sprintf "String is too long: %s/%s.",
-        length($value), $schema->{maxLength};
+      push @errors, E $path,
+        [string => maxLength => length($value), $schema->{maxLength}];
     }
   }
   if (defined $schema->{minLength}) {
     if (length($value) < $schema->{minLength}) {
-      push @errors, E $path, sprintf "String is too short: %s/%s.",
-        length($value), $schema->{minLength};
+      push @errors, E $path,
+        [string => minLength => length($value), $schema->{minLength}];
     }
   }
   if (defined $schema->{pattern}) {
     my $p = $schema->{pattern};
-    unless ($value =~ /$p/) {
-      push @errors, E $path, "String does not match '$p'";
-    }
+    push @errors, E $path, [string => pattern => $p] unless $value =~ /$p/;
   }
 
   return @errors;
@@ -1018,7 +1024,7 @@ sub _add_path_to_error_messages {
   for my $e (@errors_with_index) {
     my $index = shift @$e;
     push @errors, map {
-      my $msg = sprintf '/%s/%s %s', $type, $index, $_->{message};
+      my $msg = sprintf '/%s/%s %s', $type, $index, $_->message;
       $msg =~ s!(\d+)\s/!$1/!g;
       E $_->path, $msg;
     } @$e;
@@ -1034,12 +1040,6 @@ sub _cmp {
   return "";
 }
 
-sub _expected {
-  my $type = _guess_data_type($_[1], []);
-  return "Expected $_[0] - got different $type." if $_[0] =~ /\b$type\b/;
-  return "Expected $_[0] - got $type.";
-}
-
 # _guess_data_type($data, [{type => ...}, ...])
 sub _guess_data_type {
   my $ref     = ref $_[0];
